"""Support for reading rotation sensors

The `rotaryio` module contains classes to read different rotation encoding schemes. See
`Wikipedia's Rotary Encoder page <https://en.wikipedia.org/wiki/Rotary_encoder>`_ for more
background.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info."""

from __future__ import annotations

import microcontroller

class IncrementalEncoder:
    """IncrementalEncoder determines the relative rotational position based on two series of pulses."""

    def __init__(
        self, pin_a: microcontroller.Pin, pin_b: microcontroller.Pin, divisor: int = 4
    ) -> None:
        """Create an IncrementalEncoder object associated with the given pins. It tracks the positional
        state of an incremental rotary encoder (also known as a quadrature encoder.) Position is
        relative to the position when the object is contructed.

        :param ~microcontroller.Pin pin_a: First pin to read pulses from.
        :param ~microcontroller.Pin pin_b: Second pin to read pulses from.
        :param int divisor: The divisor of the quadrature signal.

        For example::

          import rotaryio
          import time
          from board import *

          enc = rotaryio.IncrementalEncoder(D1, D2)
          last_position = None
          while True:
              position = enc.position
              if last_position == None or position != last_position:
                  print(position)
              last_position = position"""
        ...
    def deinit(self) -> None:
        """Deinitializes the IncrementalEncoder and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> IncrementalEncoder:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    divisor: int
    """The divisor of the quadrature signal.  Use 1 for encoders without
    detents, or encoders with 4 detents per cycle.  Use 2 for encoders with 2
    detents per cycle.  Use 4 for encoders with 1 detent per cycle."""

    position: int
    """The current position in terms of pulses. The number of pulses per rotation is defined by the
    specific hardware and by the divisor."""



"""Low-level neopixel implementation

The `neopixel_write` module contains a helper method to write out bytes in
the 800khz neopixel protocol.

For example, to turn off a single neopixel (like the status pixel on Express
boards.)

.. code-block:: python

  import board
  import neopixel_write
  import digitalio

  pin = digitalio.DigitalInOut(board.NEOPIXEL)
  pin.direction = digitalio.Direction.OUTPUT
  pixel_off = bytearray([0, 0, 0])
  neopixel_write.neopixel_write(pin, pixel_off)"""

from __future__ import annotations

import digitalio
from _typing import ReadableBuffer

def neopixel_write(digitalinout: digitalio.DigitalInOut, buf: ReadableBuffer) -> None:
    """Write buf out on the given DigitalInOut.

    :param ~digitalio.DigitalInOut digitalinout: the DigitalInOut to output with
    :param ~_typing.ReadableBuffer buf: The bytes to clock out. No assumption is made about color order"""
    ...



"""Manipulation of c-style data

This module implements a subset of the corresponding CPython module,
as described below. For more information, refer to the original CPython
documentation: struct.

Supported size/byte order prefixes: *@*, *<*, *>*, *!*.

Supported format codes: *b*, *B*, *x*, *h*, *H*, *i*, *I*, *l*, *L*, *q*, *Q*,
*s*, *P*, *f*, *d* (the latter 2 depending on the floating-point support)."""

from __future__ import annotations

from typing import Any, Tuple

from _typing import ReadableBuffer, WriteableBuffer

def calcsize(fmt: str) -> int:
    """Return the number of bytes needed to store the given fmt."""
    ...

def pack(fmt: str, *values: Any) -> bytes:
    """Pack the values according to the format string fmt.
    The return value is a bytes object encoding the values."""
    ...

def pack_into(fmt: str, buffer: WriteableBuffer, offset: int, *values: Any) -> None:
    """Pack the values according to the format string fmt into a buffer
    starting at offset. offset may be negative to count from the end of buffer."""
    ...

def unpack(fmt: str, data: ReadableBuffer) -> Tuple[Any, ...]:
    """Unpack from the data according to the format string fmt. The return value
    is a tuple of the unpacked values. The buffer size must match the size
    required by the format."""
    ...

def unpack_from(fmt: str, data: ReadableBuffer, offset: int = 0) -> Tuple[Any, ...]:
    """Unpack from the data starting at offset according to the format string fmt.
    offset may be negative to count from the end of buffer. The return value is
    a tuple of the unpacked values. The buffer size must be at least as big
    as the size required by the form."""
    ...



from __future__ import annotations

import microcontroller

class PinAlarm:
    """Trigger an alarm when a pin changes state."""

    def __init__(
        self,
        pin: microcontroller.Pin,
        value: bool,
        edge: bool = False,
        pull: bool = False,
    ) -> None:
        """Create an alarm triggered by a `microcontroller.Pin` level. The alarm is not active
        until it is passed to an `alarm`-enabling function, such as `alarm.light_sleep_until_alarms()` or
        `alarm.exit_and_deep_sleep_until_alarms()`.

        :param microcontroller.Pin pin: The pin to monitor. On some ports, the choice of pin
          may be limited due to hardware restrictions, particularly for deep-sleep alarms.
        :param bool value: When active, trigger when the pin value is high (``True``) or low (``False``).
          On some ports, multiple `PinAlarm` objects may need to have coordinated values
          for deep-sleep alarms.
        :param bool edge: If ``True``, trigger only when there is a transition to the specified
          value of ``value``. If ``True``, if the alarm becomes active when the pin value already
          matches ``value``, the alarm is not triggered: the pin must transition from ``not value``
          to ``value`` to trigger the alarm. On some ports, edge-triggering may not be available,
          particularly for deep-sleep alarms.
        :param bool pull: Enable a pull-up or pull-down which pulls the pin to the level opposite
          that of ``value``. For instance, if ``value`` is set to ``True``, setting ``pull``
          to ``True`` will enable a pull-down, to hold the pin low normally until an outside signal
          pulls it high.
        """
        ...
    pin: microcontroller.Pin
    """The trigger pin."""

    value: bool
    """The value on which to trigger."""



"""Analog hardware support

The `analogio` module contains classes to provide access to analog IO
typically implemented with digital-to-analog (DAC) and analog-to-digital
(ADC) converters.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

For example::

  import analogio
  from board import *

  pin = analogio.AnalogIn(A0)
  print(pin.value)
  pin.deinit()

This example will initialize the the device, read
:py:data:`~analogio.AnalogIn.value` and then
:py:meth:`~analogio.AnalogIn.deinit` the hardware. The last step is optional
because CircuitPython will do it automatically after the program finishes."""

from __future__ import annotations

import microcontroller

class AnalogIn:
    """Read analog voltage levels

    Usage::

       import analogio
       from board import *

       adc = analogio.AnalogIn(A1)
       val = adc.value"""

    def __init__(self, pin: microcontroller.Pin) -> None:
        """Use the AnalogIn on the given pin. The reference voltage varies by
        platform so use ``reference_voltage`` to read the configured setting.

        :param ~microcontroller.Pin pin: the pin to read from"""
        ...
    def deinit(self) -> None:
        """Turn off the AnalogIn and release the pin for other use."""
        ...
    def __enter__(self) -> AnalogIn:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    value: int
    """The value on the analog pin between 0 and 65535 inclusive (16-bit). (read-only)

    Even if the underlying analog to digital converter (ADC) is lower
    resolution, the value is 16-bit."""

    reference_voltage: float
    """The maximum voltage measurable (also known as the reference voltage) as a
    `float` in Volts."""

class AnalogOut:
    """Output analog values (a specific voltage).

    Example usage::

        import analogio
        from board import *

        dac = analogio.AnalogOut(A2)                # output on pin A2
        dac.value = 32768                           # makes A2 1.65V"""

    def __init__(self, pin: microcontroller.Pin) -> None:
        """Use the AnalogOut on the given pin.

        :param ~microcontroller.Pin pin: the pin to output to"""
        ...
    def deinit(self) -> None:
        """Turn off the AnalogOut and release the pin for other use."""
        ...
    def __enter__(self) -> AnalogOut:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    value: int
    """The value on the analog pin between 0 and 65535 inclusive (16-bit). (write-only)

    Even if the underlying digital to analog converter (DAC) is lower
    resolution, the value is 16-bit."""



from __future__ import annotations

class IS31FL3741:
    """Displays an in-memory framebuffer to a IS31FL3741 drive display."""

    def __init__(self, *, width: int) -> None:
        """Create a IS31FL3741 object with the given attributes.

        The framebuffer is in "RGB888" format using 4 bytes per pixel.
        Bits 24-31 are ignored. The format is in RGB order.

        If a framebuffer is not passed in, one is allocated and initialized
        to all black.  In any case, the framebuffer can be retrieved
        by passing the Is31fl3741 object to memoryview().

        A Is31fl3741 is often used in conjunction with a
        `framebufferio.FramebufferDisplay`."""
    def deinit(self) -> None:
        """Free the resources associated with this
        IS31FL3741 instance.  After deinitialization, no further operations
        may be performed."""
        ...
    brightness: float
    """In the current implementation, 0.0 turns the display off entirely
    and any other value up to 1.0 turns the display on fully."""
    def refresh(self) -> None:
        """Transmits the color data in the buffer to the pixels so that
        they are shown."""
        ...
    width: int
    """The width of the display, in pixels"""

    height: int
    """The height of the display, in pixels"""



"""Interface to an SD card via the SPI bus"""

from __future__ import annotations

import busio
import microcontroller
from _typing import ReadableBuffer, WriteableBuffer

class SDCard:
    """SD Card Block Interface

    Controls an SD card over SPI.  This built-in module has higher read
    performance than the library adafruit_sdcard, but it is only compatible with
    `busio.SPI`, not `bitbangio.SPI`.  Usually an SDCard object is used
    with ``storage.VfsFat`` to allow file I/O to an SD card."""

    def __init__(
        self, bus: busio.SPI, cs: microcontroller.Pin, baudrate: int = 8000000
    ) -> None:
        """Construct an SPI SD Card object with the given properties

        :param busio.SPI spi: The SPI bus
        :param microcontroller.Pin cs: The chip select connected to the card
        :param int baudrate: The SPI data rate to use after card setup

        Note that during detection and configuration, a hard-coded low baudrate is used.
        Data transfers use the specified baurate (rounded down to one that is supported by
        the microcontroller)

        .. important::
           If the same SPI bus is shared with other peripherals, it is important that
           the SD card be initialized before accessing any other peripheral on the bus.
           Failure to do so can prevent the SD card from being recognized until it is
           powered off or re-inserted.

        Example usage:

        .. code-block:: python

            import os

            import board
            import sdcardio
            import storage

            sd = sdcardio.SDCard(board.SPI(), board.SD_CS)
            vfs = storage.VfsFat(sd)
            storage.mount(vfs, '/sd')
            os.listdir('/sd')"""
    def count(self) -> int:
        """Returns the total number of sectors

        Due to technical limitations, this is a function and not a property.

        :return: The number of 512-byte blocks, as a number"""
    def deinit(self) -> None:
        """Disable permanently.

        :return: None"""
    def readblocks(self, start_block: int, buf: WriteableBuffer) -> None:

        """Read one or more blocks from the card

        :param int start_block: The block to start reading from
        :param ~_typing.WriteableBuffer buf: The buffer to write into.  Length must be multiple of 512.

        :return: None"""
    def sync(self) -> None:
        """Ensure all blocks written are actually committed to the SD card

        :return: None"""
        ...
    def writeblocks(self, start_block: int, buf: ReadableBuffer) -> None:

        """Write one or more blocks to the card

        :param int start_block: The block to start writing from
        :param ~_typing.ReadableBuffer buf: The buffer to read from.  Length must be multiple of 512.

        :return: None"""



"""Support for "Parallel capture" interfaces"""

from __future__ import annotations

from typing import List, Optional

import microcontroller
from _typing import WriteableBuffer

class ParallelImageCapture:
    """Capture image frames from a camera with parallel data interface"""

    def __init__(
        self,
        *,
        data_pins: List[microcontroller.Pin],
        clock: microcontroller.Pin,
        vsync: Optional[microcontroller.Pin],
        href: Optional[microcontroller.Pin],
    ) -> None:
        """Create a parallel image capture object

        :param List[microcontroller.Pin] data_pins: The data pins.
        :param microcontroller.Pin clock: The pixel clock input.
        :param microcontroller.Pin vsync: The vertical sync input, which has a negative-going pulse at the beginning of each frame.
        :param microcontroller.Pin href: The horizontal reference input, which is high whenever the camera is transmitting valid pixel information.
        """
        ...
    def capture(self, buffer: WriteableBuffer) -> WriteableBuffer:
        """Capture a single frame into the given buffer.

        This will stop a continuous-mode capture, if one is in progress."""
        ...
    def continuous_capture_start(
        self, buffer1: WriteableBuffer, buffer2: WriteableBuffer, /
    ) -> None:
        """Begin capturing into the given buffers in the background.

        Call `continuous_capture_get_frame` to get the next available
        frame, and `continuous_capture_stop` to stop capturing.

        Until `continuous_capture_stop` (or `deinit`) is called, the
        `ParallelImageCapture` object keeps references to ``buffer1`` and
        ``buffer2``, so the objects will not be garbage collected."""
        ...
    def continuous_capture_get_frame(self) -> WriteableBuffer:
        """Return the next available frame, one of the two buffers passed to `continuous_capture_start`"""
        ...
    def continuous_capture_stop(self) -> None:
        """Stop continuous capture.

        Calling this method also causes the object to release its
        references to the buffers passed to `continuous_capture_start`,
        potentially allowing the objects to be garbage collected."""
        ...
    def deinit(self) -> None:
        """Deinitialize this instance"""
        ...
    def __enter__(self) -> ParallelImageCapture:
        """No-op used in Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware on context exit. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...



"""time and timing related functions

The `time` module is a strict subset of the CPython `cpython:time` module. So, code
using `time` written in CircuitPython will work in CPython but not necessarily the other
way around."""

from __future__ import annotations

from typing import Sequence

def monotonic() -> float:
    """Returns an always increasing value of time with an unknown reference
    point. Only use it to compare against other values from `time.monotonic()`.

    On most boards, `time.monotonic()` converts a 64-bit millisecond tick counter
    to a float. Floats on most boards are encoded in 30 bits internally, with
    effectively 22 bits of precision. The float returned by `time.monotonic()` will
    accurately represent time to millisecond precision only up to 2**22 milliseconds
    (about 1.165 hours).
    At that point it will start losing precision, and its value will change only
    every second millisecond. At 2**23 milliseconds it will change every fourth
    millisecond, and so forth.

    If you need more consistent precision, use `time.monotonic_ns()`, or `supervisor.ticks_ms()`.
    `time.monotonic_ns()` is not available on boards without long integer support.
    `supervisor.ticks_ms()` uses intervals of a millisecond, but wraps around, and is not
    CPython-compatible.

    :return: the current monotonic time
    :rtype: float"""
    ...

def sleep(seconds: float) -> None:
    """Sleep for a given number of seconds.

    :param float seconds: the time to sleep in fractional seconds"""
    ...

class struct_time:
    def __init__(self, time_tuple: Sequence[int]) -> None:
        """Structure used to capture a date and time.  Can be constructed from a `struct_time`, `tuple`, `list`, or `namedtuple` with 9 elements.

        :param Sequence time_tuple: Sequence of time info: ``(tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_wday, tm_yday, tm_isdst)``

          * ``tm_year``: the year, 2017 for example
          * ``tm_mon``: the month, range [1, 12]
          * ``tm_mday``: the day of the month, range [1, 31]
          * ``tm_hour``: the hour, range [0, 23]
          * ``tm_min``: the minute, range [0, 59]
          * ``tm_sec``: the second, range [0, 61]
          * ``tm_wday``: the day of the week, range [0, 6], Monday is 0
          * ``tm_yday``: the day of the year, range [1, 366], -1 indicates not known
          * ``tm_isdst``: 1 when in daylight savings, 0 when not, -1 if unknown."""
        ...

def time() -> int:
    """Return the current time in seconds since since Jan 1, 1970.

    :return: the current time
    :rtype: int"""
    ...

def monotonic_ns() -> int:
    """Return the time of the monotonic clock, which cannot go backward, in nanoseconds.
    Not available on boards without long integer support.

    :return: the current time
    :rtype: int"""
    ...

def localtime(secs: int) -> struct_time:
    """Convert a time expressed in seconds since Jan 1, 1970 to a struct_time in
    local time. If secs is not provided or None, the current time as returned
    by time() is used.
    The earliest date for which it can generate a time is Jan 1, 2000.

    :return: the current time
    :rtype: time.struct_time"""
    ...

def mktime(t: struct_time) -> int:
    """This is the inverse function of localtime(). Its argument is the
    struct_time or full 9-tuple (since the dst flag is needed; use -1 as the
    dst flag if it is unknown) which expresses the time in local time, not UTC.
    The earliest date for which it can generate a time is Jan 1, 2000.

    :return: seconds
    :rtype: int"""
    ...
